"""
Utilities to help with catalogs.
"""

import fnmatch
import pathlib
import re

import cf_xarray  # noqa
import intake
import numpy as np
import pandas as pd
import yaml

from siphon.catalog import TDSCatalog

import model_catalogs as mc


def astype(value, type_):
    """Return string or list as list"""
    if not isinstance(value, type_):
        if type_ == list and isinstance(value, (str, pathlib.PosixPath)):
            return [value]
        return type_(value)
    return value


def get_fresh_parameter(filename):
    """Get freshness parameter, based on the filename.

    A freshness parameter is stored in `__init__` for required scenarios which is looked up using the
    logic in this function, based on the filename.

    Parameters
    ----------
    filename : Path
        Filename to determine freshness

    Returns
    -------
    mu, a pandas Timedelta-interpretable string describing the amount of time that filename should be
    considered fresh before needing to be recalculated.
    """

    # a start or end datetime file
    if filename.parent == mc.CACHE_PATH_AVAILABILITY:
        timing = filename.name.split("_")[1]
        if "start" in filename.name:
            mu = mc.FRESH[timing]["start"]
        elif "end" in filename.name:
            mu = mc.FRESH[timing]["end"]
        elif "catrefs" in filename.name:
            mu = mc.FRESH[timing]["catrefs"]
    # a file of file locs for aggregation
    elif filename.parent == mc.CACHE_PATH_FILE_LOCS:
        timing = filename.name.split("_")[1]
        mu = mc.FRESH[timing]["file_locs"]
    # a compiled catalog file
    elif filename.parent == mc.CACHE_PATH_COMPILED:
        mu = mc.FRESH["compiled"]

    return mu


def is_fresh(filename):
    """Check if file called filename is fresh.

    If filename doesn't exist, return False.

    Parameters
    ----------
    filename : Path
        Filename to determine freshness

    Returns
    -------
    Boolean. True if fresh and False if not or if filename is not found.
    """

    now = pd.Timestamp.today(tz="UTC")
    try:
        filetime = pd.Timestamp(filename.stat().st_mtime_ns).tz_localize("UTC")

        mu = get_fresh_parameter(filename)

        return now - filetime < pd.Timedelta(mu)

    except FileNotFoundError:
        return False


def find_bbox(ds, dd=None, alpha=None):
    """Determine bounds and boundary of model.

    Parameters
    ----------
    ds: Dataset
        xarray Dataset containing model output.
    dd: int, optional
        Number to decimate model output lon/lat by, as a stride.
    alpha: float, optional
        Number for alphashape to determine what counts as the convex hull.
        Larger number is more detailed, 1 is a good starting point.

    Returns
    -------
    List containing the name of the longitude and latitude variables for ds,
    geographic bounding box of model output: [min_lon, min_lat, max_lon,
    max_lat], low res and high res wkt representation of model boundary.
    """

    import shapely.geometry

    hasmask = False

    try:
        lon = ds.cf["longitude"].values
        lat = ds.cf["latitude"].values
        lonkey = ds.cf["longitude"].name
        latkey = ds.cf["latitude"].name

    except KeyError:
        if "lon_rho" in ds:
            lonkey = "lon_rho"
            latkey = "lat_rho"
        else:
            lonkey = list(ds.cf[["longitude"]].coords.keys())[0]
            # need to make sure latkey matches lonkey grid
            latkey = f"lat{lonkey[3:]}"
        # In case there are multiple grids, just take first one;
        # they are close enough
        lon = ds[lonkey].values
        lat = ds[latkey].values

    # check for corresponding mask (rectilinear and curvilinear grids)
    if any([var for var in ds.data_vars if "mask" in var]):
        if ("mask_rho" in ds) and (lonkey == "lon_rho"):
            maskkey = lonkey.replace("lon", "mask")
        elif "mask" in ds:
            maskkey = "mask"
        else:
            maskkey = None
        if maskkey in ds:
            lon = ds[lonkey].where(ds[maskkey] == 1).values
            lon = lon[~np.isnan(lon)].flatten()
            lat = ds[latkey].where(ds[maskkey] == 1).values
            lat = lat[~np.isnan(lat)].flatten()
            hasmask = True

    # This is structured, rectilinear
    # GFS, RTOFS, HYCOM
    if (lon.ndim == 1) and ("nele" not in ds.dims) and not hasmask:
        nlon, nlat = ds["lon"].size, ds["lat"].size
        lonb = np.concatenate(([lon[0]] * nlat, lon[:], [lon[-1]] * nlat, lon[::-1]))
        latb = np.concatenate((lat[:], [lat[-1]] * nlon, lat[::-1], [lat[0]] * nlon))
        # boundary = np.vstack((lonb, latb)).T
        p = shapely.geometry.Polygon(zip(lonb, latb))
        p0 = p.simplify(1)
        # Now using the more simplified version because all of these models are boxes
        p1 = p0

    elif hasmask or ("nele" in ds.dims):  # unstructured

        assertion = (
            "dd and alpha need to be defined in the source_catalog for this model."
        )
        assert dd is not None and alpha is not None, assertion

        # need to calculate concave hull or alphashape of grid
        import alphashape

        # low res, same as convex hull
        p0 = alphashape.alphashape(list(zip(lon, lat)), 0.0)
        # downsample a bit to save time, still should clearly see shape of domain
        pts = shapely.geometry.MultiPoint(list(zip(lon[::dd], lat[::dd])))
        p1 = alphashape.alphashape(pts, alpha)

    # useful things to look at: p.wkt  #shapely.geometry.mapping(p)
    return lonkey, latkey, list(p0.bounds), p1.wkt
    # return lonkey, latkey, list(p0.bounds), p0.wkt, p1.wkt


def agg_for_date(date, strings, filetype, is_forecast=False, pattern=None):
    """Aggregate NOAA OFS-style nowcast/forecast files.

    Parameters
    ----------
    date: str of datetime, pd.Timestamp
        Date of day to find model output files for. Doesn't pay attention to hours/minutes seconds.
    strings: list
        List of strings to be filtered. Expected to be file locations from a thredds catalog.
    filetype: str
        Which filetype to use. Every NOAA OFS model has "fields" available, but some have "regulargrid"
        or "2ds" also. This availability information is in the source catalog for the model under
        `filetypes` metadata.
    is_forecast: bool, optional
        If True, then date is the last day of the time period being sought and the forecast files should
        be brought in along with the nowcast files, to get the model output the length of the forecast
        out in time. The forecast files brought in will have the latest timing cycle of the day that is
        available. If False, all nowcast files (for all timing cycles) are brought in.
    pattern: str, optional
        If a model file pattern doesn't match that assumed in this code, input one that will work.
        Currently only NYOFS doesn't match but the pattern is built into the catalog file.

    Returns
    -------
    List of URLs for where to find all of the model output files that match the keyword arguments.
    """

    date = astype(date, pd.Timestamp)

    # # brings in nowcast and forecast for any date in the catalog
    # pattern0 = f'*{filetype}*.t??z.*'
    # # brings in nowcast and forecast for only the day specified
    # pattern0 = date.strftime(f'*{filetype}*.%Y%m%d.t??z.*')

    if pattern is None:
        pattern = date.strftime(f"*{filetype}*.n*.%Y%m%d.t??z.*")
        # pattern = date.strftime(f'*{filetype}*.n???.%Y%m%d.t??z.*')
    else:
        pattern = eval(f"f'{pattern}'")
    # pattern = date.strftime(f'*{filetype}*.n*.%Y%m%d.t??z.*')

    # '*{filetype}.n???.{date.year}{str(date.month).zfill(2)}{str(date.day).zfill(2)}.t??z.*'

    # pattern = eval(f"f'{pattern}'")
    # import pdb; pdb.set_trace()
    fnames = fnmatch.filter(strings, pattern)

    if is_forecast:

        import re

        regex = re.compile(".t[0-9]{2}z.")
        # substrings
        subs = [substr[2:4] for substr in regex.findall("".join(strings))]
        cycle = sorted(list(set(subs)))[-1]  # noqa: F841

        # cycle = sorted(list(set([fname[ fname.find(start:='.t') + len(start):fname.find('z.')] for fname in fnames])))[-1]  # noqa: E501
        # import pdb; pdb.set_trace()
        # pattern1 = f'*{filetype}*.t{cycle}z.*'

        # replace '.t??z.' in pattern with '.t{cycle}z.' and replace '.n*.' with '.*.'
        pattern1 = pattern.replace(".t??z.", ".t{cycle}z.").replace(".n*.", ".*.")
        pattern1 = eval(f"f'{pattern1}'")
        fnames = fnmatch.filter(strings, pattern1)

    # filelocs = [cat.datasets.get(fname).access_urls["OPENDAP"] for fname in fnames]

    return fnames


def find_catrefs(catloc):
    """Find hierarchy of catalog references for thredds catalog.

    Parameters
    ----------
    catloc: str
        Search in thredds catalog structure from base catalog, catloc.

    Returns
    -------
    list of tuples containing the hierarchy of directories in the thredds catalog
    structure to get to where the datafiles start.
    """

    # 0th level catalog
    cat = TDSCatalog(catloc)
    catrefs_to_check = list(cat.catalog_refs)
    # only keep numerical directories
    catrefs_to_check = [catref for catref in catrefs_to_check if catref.isnumeric()]

    # 1st level catalog
    catrefs_to_check1 = [
        cat.catalog_refs[catref].follow().catalog_refs for catref in catrefs_to_check
    ]

    # Combine the catalog references together
    catrefs = [
        (catref0, catref1)
        for catref0, catrefs1 in zip(catrefs_to_check, catrefs_to_check1)
        for catref1 in catrefs1
    ]

    # Check first one to see if there are more catalog references or not
    cat_ref_test = (
        cat.catalog_refs[catrefs[0][0]]
        .follow()
        .catalog_refs[catrefs[0][1]]
        .follow()
        .catalog_refs
    )
    # If there are more catalog references, run another level of catalog and combine, ## 2
    if len(cat_ref_test) > 1:
        catrefs2 = [
            cat.catalog_refs[catref[0]]
            .follow()
            .catalog_refs[catref[1]]
            .follow()
            .catalog_refs
            for catref in catrefs
        ]
        catrefs = [
            (catref[0], catref[1], catref22)
            for catref, catref2 in zip(catrefs, catrefs2)
            for catref22 in catref2
        ]

    # OUTPUT
    return catrefs


def find_filelocs(catref, catloc, filetype="fields"):
    """Find thredds file locations.

    Parameters
    ----------
    catref: tuple
        2 or 3 labels describing the directories from catlog to get the data
        locations.
    catloc: str
        Base thredds catalog location.
    filetype: str
        Which filetype to use. Every NOAA OFS model has "fields" available, but
        some have "regulargrid"or "2ds" also (listed in separate catalogs in the
        model name).

    Returns
    -------
    Locations of files found from catloc to hierarchical location described by
    catref.
    """

    filelocs = []
    cat = TDSCatalog(catloc)
    if len(catref) == 2:
        catref1, catref2 = catref
        cat1 = cat.catalog_refs[catref1].follow()
        cat2 = cat1.catalog_refs[catref2].follow()
        last_cat = cat2
        datasets = cat2.datasets
    elif len(catref) == 3:
        catref1, catref2, catref3 = catref
        cat1 = cat.catalog_refs[catref1].follow()
        cat2 = cat1.catalog_refs[catref2].follow()
        cat3 = cat2.catalog_refs[catref3].follow()
        last_cat = cat3
        datasets = cat3.datasets

    for dataset in datasets:
        if (
            "stations" not in dataset
            and "vibrioprob" not in dataset
            and filetype in dataset
        ):

            url = last_cat.datasets[dataset].access_urls["OPENDAP"]
            filelocs.append(url)
    return filelocs


def get_dates_from_ofs(filelocs, filetype, norf, firstorlast):
    """Return either start or end datetime from list of filenames.

    This looks at the actual nowcast and forecast file cycle times to understand the earliest and last
    model times, as opposed to just the date in the file name.

    Parameters
    ----------
    filelocs: list
        Locations of files found from catloc to hierarchical location described by
        catref.
    filetype: str
        Which filetype to use. Every NOAA OFS model has "fields" available, but
        some have "regulargrid"or "2ds" also (listed in separate catalogs in the
        model name).
    norf: str
        "n" or "f" for "nowcast" or "forecast", for OFS files.
    firstorlast: str
        Whether to get the "first" or "last" entry of the filelocs, which will
        translate to the index to use.
    """

    pattern = f"*{filetype}*.{norf}*.????????.t??z.*"
    # pattern = date.strftime(f"*{filetype}*.n*.%Y%m%d.t??z.*")
    if firstorlast == "first":
        ind = 0
    elif firstorlast == "last":
        ind = -1
    fileloc = sorted(fnmatch.filter(filelocs, pattern))[ind]
    filename = fileloc.split("/")[-1]
    date = pd.Timestamp(filename.split(".")[4])
    regex = re.compile(".t[0-9]{2}z.")
    cycle = [substr[2:4] for substr in regex.findall("".join(filename))][0]
    regex = re.compile(f".{norf}" + "[0-9]{3}.")
    try:
        filetime = [substr[2:5] for substr in regex.findall("".join(filename))][0]
    except IndexError:  # NYOFS uses this
        filetime = 0
    # in UTC
    datetime = date + pd.Timedelta(f"{cycle} hours") + pd.Timedelta(f"{filetime} hours")

    return datetime


def calculate_boundaries(file_locs=None, save_files=True, return_boundaries=False):
    """Calculate boundary information for all models.

    This loops over all catalog files available in mc.CAT_PATH_ORIG, tries first with forecast source and
    then with nowcast source if necessary to access the example model output files and calculate the
    bounding box and numerical domain boundary. The numerical domain boundary is calculated using
    `alpha_shape` with previously-chosen parameters stored in the original model catalog files. The
    bounding box and boundary string representation (as WKT) are then saved to files.

    The files that are saved by running this function have been previously saved into the repository, so
    this function should only be run if you suspect that a model domain has changed.

    Parameters
    ----------
    file_locs : Path, list of Paths, optional
        List of Path objects for model catalog files to read from. If not input, will use all catalog
        files available at mc.CAT_PATH_ORIG.glob("*.yaml").
    save_files : boolean, optional
        Whether to save files or not. Defaults to True. Saves to mc.CAT_PATH_BOUNDARY / cat_loc.name.
    return_boundaries : boolean, optional
        Whether to return boundaries information from this call. Defaults to False.

    Examples
    --------
    Calculate boundary information for all available models:
    >>> mc.calculate_boundaries()

    Calculate boundary information for CBOFS:
    >>> mc.calculate_boundaries([mc.CAT_PATH_ORIG / "cbofs.yaml"])
    """

    if file_locs is None:
        file_locs = mc.CAT_PATH_ORIG.glob("*.yaml")
    else:
        file_locs = mc.astype(file_locs, list)

    # loop over all orig catalogs
    boundaries = {}
    for cat_loc in file_locs:

        # open model catalog
        cat_orig = intake.open_catalog(cat_loc)

        # try with forecast but if it doesn't work, use nowcast
        # this avoids problematic NOAA OFS aggregations when they are broken
        try:
            timing = "forecast"
            source_orig = cat_orig[timing]
            source_transform = mc.transform_source(source_orig)

            # need to make catalog to transfer information properly from
            # source_orig to source_transform
            cat_transform = mc.make_catalog(
                source_transform,
                full_cat_name=cat_orig.name,  # model name
                full_cat_description=cat_orig.description,
                full_cat_metadata=cat_orig.metadata,
                cat_driver=mc.process.DatasetTransform,
                cat_path=None,
                save_catalog=False,
            )

            # read in model output
            ds = cat_transform[timing].to_dask()

        except OSError:
            timing = "nowcast"
            source_orig = cat_orig[timing]
            source_transform = mc.transform_source(source_orig)

            # need to make catalog to transfer information properly from
            # source_orig to source_transform
            cat_transform = mc.make_catalog(
                source_transform,
                full_cat_name=cat_orig.name,  # model name
                full_cat_description=cat_orig.description,
                full_cat_metadata=cat_orig.metadata,
                cat_driver=mc.process.DatasetTransform,
                cat_path=None,
                save_catalog=False,
            )

            # read in model output
            ds = cat_transform[timing].to_dask()

        # find boundary information for model
        if "alpha_shape" in cat_orig.metadata:
            dd, alpha = cat_orig.metadata["alpha_shape"]
        else:
            dd, alpha = None, None
        lonkey, latkey, bbox, wkt = mc.find_bbox(ds, dd=dd, alpha=alpha)

        ds.close()

        # save boundary info to file
        if save_files:
            with open(mc.FILE_PATH_BOUNDARIES(cat_loc.name), "w") as outfile:
                yaml.dump({"bbox": bbox, "wkt": wkt}, outfile, default_flow_style=False)

        if return_boundaries:
            boundaries[cat_loc.stem] = {"bbox": bbox, "wkt": wkt}

    if return_boundaries:
        return boundaries



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""A module for common components for scripting examples."""
import time

from argparse import ArgumentParser
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Callable, List, Tuple

import pandas as pd
import xarray as xr

from extract_model import utils as em_utils

import model_catalogs as mc


class Timer:
    """A class to aid with measuring timing."""

    def __init__(self, msg=None):
        """Initializes the current time."""
        self.t0 = time.time()
        self.msg = msg

    def tick(self):
        """Update the timer start."""
        self.t0 = time.time()

    def tock(self) -> float:
        """Return elapsed time in ms."""
        return (time.time() - self.t0) * 1000.0

    def format(self):
        time_in_ms = self.tock()
        if time_in_ms > 60000:
            return f"{time_in_ms / 60000:.1f} min"
        if time_in_ms > 2000:
            return f"{time_in_ms / 1000:.1f} s"
        return f"{time_in_ms:.1f} ms"

    def __enter__(self):
        """With context."""
        self.tick()

    def __exit__(self, type, value, traceback):
        """With exit."""
        if self.msg is not None:
            print(self.msg.format(self.format()))


STANDARD_NAMES = [
    "eastward_sea_water_velocity",
    "northward_sea_water_velocity",
    "eastward_wind",
    "northward_wind",
    "sea_water_temperature",
    "sea_water_practical_salinity",
    "sea_floor_depth",
]


@dataclass
class FetchConfig:
    """Configuration data class for fetching."""

    model_name: str
    output_pth: Path
    start: pd.Timestamp
    end: pd.Timestamp
    bbox: Tuple[float, float, float, float]
    timing: str
    standard_names: List[str] = field(default_factory=lambda: STANDARD_NAMES)
    surface_only: bool = False


def parse_bbox(val: str) -> Tuple[float, float, float, float]:
    """Return the bounding box parsed from the comma-delimited string.

    Parameters
    ----------
    val : str
        Comma-delimited sequence of 4 float values for (lon_min, lat_min, lon_max, lat_max)

    Returns
    -------
        tuple of 4 floats
    """
    values = val.split(",")
    if len(values) != 4:
        raise ValueError(
            "bbox should include four numbers: lon_min,lat_min,lon_max,lat_max"
        )
    return tuple(float(i) for i in values)


def get_surface(ds: xr.Dataset) -> xr.Dataset:
    """Return a dataset that is reduced to only the surface layer."""
    model_guess = em_utils.guess_model_type(ds)
    if all([ds[zaxis].ndim < 2 for zaxis in ds.cf.axes["Z"]]):
        return ds.cf.sel(Z=0, method="nearest")
    elif model_guess == "FVCOM":
        vertical_dims = set()
        for varname in ds.cf.axes["Z"]:
            vertical_dim = ds[varname].dims[0]
            vertical_dims.add(vertical_dim)
        sel_kwargs = {vdim: 0 for vdim in vertical_dims}
        return ds.isel(**sel_kwargs)
    elif model_guess == "SELFE":
        return ds.isel(nv=-1)
    raise ValueError("Can't decode vertical coordinates.")


def fetch(fetch_config: FetchConfig):
    """Downloads and subsets the model data.

    Parameters
    ----------
    fetch_config : FetchConfig
        The configuration object which contains the model name, timing, start/end dates of the
        request, etc.
    """
    print("Setting up source catalog")
    with Timer("\tSource catalog generated in {}"):
        main_cat = mc.setup()

    print(
        f"Generating catalog specific for {fetch_config.model_name} {fetch_config.timing}"
    )
    with Timer("\tSpecific catalog generated in {}"):
        source = mc.select_date_range(
            main_cat[fetch_config.model_name],
            start_date=fetch_config.start,
            end_date=fetch_config.end,
            timing=fetch_config.timing,
        )
    print("Getting xarray dataset for model data")
    with Timer("\tCreated dask-based xarray dataset in {}"):
        ds = source.to_dask()

    if fetch_config.surface_only:
        print("Selecting only surface data.")
        with Timer("\tIndexed surface data in {}"):
            ds = get_surface(ds)
    print("Subsetting data")
    with Timer("\tSubsetted dataset in {}"):
        ds_ss = (
            ds.em.filter(fetch_config.standard_names)
            # .cf.sel(T=slice(fetch_config.start, fetch_config.end))
            .em.sub_grid(bbox=fetch_config.bbox)
        )
    print(
        f"Writing netCDF data to {fetch_config.output_pth}. This may take a long time..."
    )
    with Timer("\tWrote output to disk in {}"):
        ds_ss.to_netcdf(fetch_config.output_pth)
    print("Complete")


def parse_config(
    main: Callable,
    model_name: str,
    default_bbox: Tuple[float, float, float, float],
    output_dir: Path,
    default_timing: str = "hindcast",
    standard_names: List[str] = None,
    default_start: datetime = None,
    default_end: datetime = None,
) -> FetchConfig:
    """Parse command line arguments into a FetchConfig object.

    Parameters
    ----------
    main : function
        A reference to the main function of the script. This is used to fill in the help output
        while parsing arguments.
    model_name : str
        Name of the model (as it appears in the catalog files).
    default_bbox : tuple of floats
        The default bounding box to use for subsetting if the user does not specify one in the
        command line arguments.
    output_dir : Path
        The output path for where to write resulting netCDF files to.
    default_timing : str
        The default model run-type to use if not specified by CLI arguments. One of "forecast",
        "nowcast", or "hindcast".
    standard_names : list of strings
        The default list of standard names to use to filter on if not specified by CLI arguments.
    default_start : datetime
        The default start time of the query to use if not specified by CLI arguments.
    default_end : datetime
        The default end time of the query to use if not specified by CLI arguments.

    Returns
    -------
    FetchConfig
        An object which contains all of the information needed by the `fetch` function for
        requesting, subsetting, and filtering a dataset.

    """
    if standard_names is None:
        standard_names = STANDARD_NAMES
    parser = ArgumentParser(description=main.__doc__)
    parser.add_argument(
        "-t",
        "--timing",
        choices=["hindcast", "nowcast", "forecast"],
        default=default_timing,
        help="Model Timing Choice.",
    )
    parser.add_argument(
        "-s",
        "--start",
        type=pd.Timestamp,
        default=default_start,
        help="Request start time",
    )
    parser.add_argument(
        "-e", "--end", type=pd.Timestamp, default=default_end, help="Request end time"
    )
    parser.add_argument(
        "-f", "--force", action="store_true", help="Overwrite existing files"
    )
    parser.add_argument(
        "--bbox", type=parse_bbox, default=default_bbox, help="Specify the bounding box"
    )
    parser.add_argument(
        "--surface", action="store_true", default=False, help="Fetch only surface data."
    )
    args = parser.parse_args()

    # Sanity check on start/end time
    if args.start is None and args.end is None:
        start = pd.Timestamp("2022-06-20")
        end = pd.Timestamp("2022-06-21")
    elif args.start is None or args.end is None:
        raise ValueError("start time or end time not specified")
    else:
        start = args.start
        end = args.end
    if start >= end:
        raise ValueError("end time must be greater than start time")

    output_filename = f"{model_name}_{args.timing}_{start:%Y%m%d}-{end:%Y%m%d}.nc"
    output_pth = output_dir / output_filename
    if output_pth.exists():
        if args.force:
            output_pth.unlink()
        else:
            raise FileExistsError(f"{output_pth} already exists")

    return FetchConfig(
        model_name=model_name,
        output_pth=output_pth,
        start=start,
        end=end,
        bbox=args.bbox,
        timing=args.timing,
        standard_names=standard_names,
        surface_only=args.surface,
    )



#!/usr/bin/env pytest
# -*- coding: utf-8 -*-
"""Unit tests for example support module."""
from model_catalogs.examples import parse_bbox


def test_bbox_parse():
    """Test that bbox can be parsed."""
    bbox = parse_bbox("1,2,3,4")
    assert bbox == (1.0, 2.0, 3.0, 4.0)



"""
Everything dealing with the catalogs.
"""

import warnings

from pathlib import Path

import cf_xarray  # noqa
import intake
import intake.source.derived
import pandas as pd
import yaml

from datetimerange import DateTimeRange
from intake.catalog import Catalog
from intake.catalog.local import LocalCatalogEntry

import model_catalogs as mc


def make_catalog(
    cats,
    full_cat_name,
    full_cat_description,
    full_cat_metadata,
    cat_driver,
    cat_path=None,
    save_catalog=True,
):
    """Construct single catalog from multiple catalogs or sources.

    Parameters
    ----------
    cats: list
       List of Intake catalog or source objects that will be combined into a single catalog.
    full_cat_name: str
       Name of overall catalog.
    full_cat_descrption: str
       Description of overall catalog.
    full_cat_metadata: dict
       Dictionary of metadata for overall catalog.
    cat_driver: str or Intake object or list
       Driver to apply to all catalog entries. For example:
       * intake.catalog.local.YAMLFileCatalog
       * 'opendap'
       If list, must be same length as cats and contains drivers that
       correspond to cats.
    cat_path: Path object, optional
       Path with catalog name to use for saving catalog. With or without yaml suffix. If not provided,
       will use `full_cat_name`.
    save_catalog : bool, optional
        Defaults to True, and saves to cat_path.

    Returns
    -------

    Intake catalog.

    Examples
    --------

    Make catalog:

    >>> make_catalog([list of Intake sources or catalogs], 'catalog name', 'catalog desc', {}, 'opendap',
    save_catalog=False)
    """

    if cat_path is None and save_catalog:
        cat_path = Path(full_cat_name)
    if save_catalog:
        cat_path = cat_path / full_cat_name.lower()
        # cat_path = f"{cat_path}/{full_cat_name.lower()}"
    if save_catalog and ("yaml" not in str(cat_path)) and ("yml" not in str(cat_path)):
        cat_path = cat_path.with_suffix(".yaml")

    if not isinstance(cats, list):
        cats = [cats]
    if not isinstance(cat_driver, list):
        cat_driver = [cat_driver] * len(cats)
    assert len(cat_driver) == len(
        cats
    ), "Number of catalogs and catalog drivers must match"

    # create dictionary of catalog entries
    entries = {
        cat.name: LocalCatalogEntry(
            cat.name.upper() if isinstance(cat, Catalog) else cat.name,
            description=cat.description,
            driver=catd,
            args=cat._yaml()["sources"][cat.name]["args"],
            metadata=cat.metadata,
        )
        for cat, catd in zip(cats, cat_driver)
    }

    # create catalog
    cat = Catalog.from_dict(
        entries,
        name=full_cat_name,
        description=full_cat_description,
        metadata=full_cat_metadata,
    )

    # save catalog
    if save_catalog:
        cat.save(cat_path)

    return cat


def setup(override=False):
    """Setup reference catalogs for models.

    Loops over hard-wired "orig" catalogs available in mc.CAT_PATH_ORIG, reads in previously-saved model
    boundary information, saves temporary catalog files for each model, and links those together into the
    returned master catalog. For some models, reading in the original catalogs applies a "today" and/or
    "yesterday" date Intake user parameter that supplies two example model files that can be used for
    examining the model output for the example times. Those are rerun each time this function is rerun,
    filling the parameters using the proper dates.

    Parameters
    ----------
    override : boolean, optional
        Use `override=True` to compile the catalog files together regardless of freshness.

    Returns
    -------
    Nested Intake catalog with a source for each model in mc.CAT_PATH_ORIG. Each source/model in turn has
    a source for each timing available (e.g., "forecast", "hindcast").

    Examples
    --------
    Set up master catalog:
    >>> cat = mc.setup()

    Examine list of sources/models available in catalog:
    >>> list(cat)

    Examine the sources for a specific model in the catalog:
    >>> list(cat['CBOFS'])
    """

    cat_transform_locs = []
    # Loop over all hard-wired original catalog files, one per model
    for cat_loc in mc.CAT_PATH_ORIG.glob("*.yaml"):

        # re-compile together catalog file if user wants to override possibly
        # existing file or if is not fresh
        if override or not mc.is_fresh(mc.FILE_PATH_COMPILED(cat_loc.stem)):

            cat_orig = intake.open_catalog(cat_loc)

            # add previously-saved boundary info
            # this was calculated with mc.calculate_boundaries()
            with open(mc.FILE_PATH_BOUNDARIES(cat_loc.stem), "r") as stream:
                boundary = yaml.safe_load(stream)
            # add to cat_orig metadata
            cat_orig.metadata["bounding_box"] = boundary["bbox"]
            cat_orig.metadata["geospatial_bounds"] = boundary["wkt"]

            # get transform of each original catalog file, which points to
            # original file but applies metadata from original catalog file
            # to the resulting dataset after calling `to_dask()`
            source_transforms = [
                mc.transform_source(cat_orig[timing]) for timing in list(cat_orig)
            ]

            # need to make catalog to transfer information properly from
            # source_orig to source_transform
            mc.make_catalog(
                source_transforms,
                full_cat_name=cat_orig.name,  # model name
                full_cat_description=cat_orig.description,
                full_cat_metadata=cat_orig.metadata,
                cat_driver=mc.process.DatasetTransform,
                cat_path=mc.CACHE_PATH_COMPILED,
                save_catalog=True,
            )
        cat_transform_locs.append(mc.FILE_PATH_COMPILED(cat_loc.stem))

    # have to read these from disk in order to make them type
    # intake.catalog.local.YAMLFileCatalog
    # instead of intake.catalog.base.Catalog
    cats = [intake.open_catalog(loc) for loc in cat_transform_locs]

    # make master nested catalog
    cat = mc.make_catalog(
        cats,
        full_cat_name="MAIN-CATALOG",
        full_cat_description="Main catalog for models; a catalog of nested catalogs.",
        full_cat_metadata={"source_catalog_dir": str(mc.CAT_PATH_ORIG)},
        cat_driver=intake.catalog.local.YAMLFileCatalog,
        cat_path=None,
        save_catalog=False,
    )

    return cat


def find_datetimes(source, find_start_datetime, find_end_datetime, override=False):
    """Find the start and/or end datetimes for source.

    For sources with static urlpaths, this opens the Dataset and checks the first time for start_datetime
    and the last time for end_datetime. Some NOAA OFS models require aggregation: some forecasts, all
    nowcasts, and all hindcasts. For these, the available year and months of the thredd server
    subcatalogs are found with `find_catrefs()`. `start_datetime` is found by further evaluating to make
    sure that files in the subcatalogs are both available on the page and that the days represented by
    model output files are consecutive (there are missing dates). `end_datetime` is found from the most
    recent subcatalog files since there aren't missing files and dates on the recent end of the time
    ranges.

    Uses `cf-xarray` to determine the time axis.

    Parameters
    ----------
    source : Intake source
        Model source for which to find start and/or end datetimes
    find_start_datetime : bool
        True to calculate start_datetime, otherwise returns None
    find_start_datetime : bool
        True to calculate start_datetime, otherwise returns None
    override : boolean, optional
        Use `override=True` to find catrefs regardless of freshness. This is passed in from
        `find_availability()` so has the same value as input there.

    Returns
    -------
    (start_datetime, end_datetime) where each are strings or can be None if they didn't need to be found.
    """

    filetype = source.cat.metadata["filetype"]

    # For any model/timing pairs with static links or known file address,
    # which is all non-OFS models and OFS models that are already aggregated
    if "catloc" not in source.metadata:

        try:
            ds = source.to_dask()
            # import pdb; pdb.set_trace()
            # use one T in case there are more than one
            start_datetime = (
                str(ds[ds.cf.axes["T"][0]].values[0]) if find_start_datetime else None
            )
            end_datetime = (
                str(ds[ds.cf.axes["T"][0]].values[-1]) if find_end_datetime else None
            )
            ds.close()
        except OSError:
            warnings.warn(
                f"Model {source.cat.name} with timing {source.name} cannot connect to server.",
                RuntimeWarning,
            )
            return None, None

    # for when we need to aggregate which is OFS models nowcast and hindcast
    # and forecast if there is no pre-made aggregation
    else:
        if not override and mc.is_fresh(
            mc.FILE_PATH_CATREFS(source.cat.name, source.name)
        ):
            with open(
                mc.FILE_PATH_CATREFS(source.cat.name, source.name), "r"
            ) as stream:
                catrefs = yaml.safe_load(stream)["catrefs"]
        else:
            catrefs = mc.find_catrefs(source.metadata["catloc"])
            catrefs = sorted(catrefs)  # earliest first, most recent last
            with open(
                mc.FILE_PATH_CATREFS(source.cat.name, source.name), "w"
            ) as outfile:
                yaml.dump({"catrefs": catrefs}, outfile, default_flow_style=False)

        if find_start_datetime:
            # Getting start date #
            # first make sure the dates actually have model files available
            filelocs = []
            i = 0
            while len(filelocs) == 0:
                # print(catrefs[i])
                filelocs = sorted(
                    mc.find_filelocs(
                        catrefs[i], source.metadata["catloc"], filetype=filetype
                    )
                )
                i += 1

            # second make sure we only count when dates are consecutive, since servers tend to have some
            # spotty model output at the earliest dates get dates from file names
            all_dates = [
                pd.to_datetime(fileloc, format="%Y%m%d", exact=False)
                for fileloc in filelocs
            ]
            unique_dates = list(set(all_dates))

            # if any dates are not consecutive, need to start after that date
            df = pd.Series(unique_dates)
            ddf = df.diff() > pd.Timedelta(
                "1d"
            )  # which differences in consecutive dates are over 1 day
            if ddf.any():
                # first date after last jump in dates is desired start day
                start_day = df.where(ddf).dropna().iloc[-1]
                # subset filelocs to match discovered consecutive dates
                df_filelocs = pd.Series(index=all_dates, data=filelocs)
                filelocs_ss = list(
                    df_filelocs.where(df_filelocs.index >= start_day).dropna().values
                )

                # want first nowcast file (no forecast files available)
                start_datetime = str(
                    mc.get_dates_from_ofs(filelocs_ss, filetype, "n", "first")
                )

            # all dates were fine, so just use earliest fileloc
            else:
                # running the following gives the actual first time. This might not be necessary in which
                # case want first nowcast file (no forecast files available)
                start_datetime = str(
                    mc.get_dates_from_ofs(filelocs, filetype, "n", "first")
                )
                # # just use earliest day date
                # start_datetime = df.iloc[0]
        else:
            start_datetime = None

        if find_end_datetime:
            # Getting end date #
            filelocs = mc.find_filelocs(
                catrefs[-1], source.metadata["catloc"], filetype=filetype
            )
            # want last file
            if source.name == "hindcast":
                norf = "n"
            elif source.name in ("nowcast", "forecast"):
                norf = "f"
            end_datetime = str(mc.get_dates_from_ofs(filelocs, filetype, norf, "last"))
        else:
            end_datetime = None

    # save start/end to files
    if find_start_datetime:
        with open(mc.FILE_PATH_START(source.cat.name, source.name), "w") as outfile:
            yaml.dump(
                {"start_datetime": start_datetime}, outfile, default_flow_style=False
            )
    if find_end_datetime:
        with open(mc.FILE_PATH_END(source.cat.name, source.name), "w") as outfile:
            yaml.dump({"end_datetime": end_datetime}, outfile, default_flow_style=False)

    return start_datetime, end_datetime


def find_availability(cat, timings=None, override=False):
    """Find availability for model timings.

    The code will check for previously-calculated availability. If found, the "freshness" of the
    information is checked as compared with mc.FRESH parameters specified in `__init__`.

    Start and end datetimes are allowed to be calculated separately to save time.

    Parameters
    ----------
    cat : Intake catalog
        Catalog containing timing sources for which to find availability.
    timings : str, list of strings, optional
        Specified timing to find the availability for. If unspecified, loop over all timings and find
        availability for all.
    override : boolean, optional
        Use `override=True` to find availability regardless of freshness.

    Returns
    -------
    The input Intake catalog but with `start_datetime` and `end_datetime` added to metadata for the
    timings that were evaluated.

    Examples
    --------
    Setup source catalog, then find availability for all timings of CIOFS model:
    >>> source_cat = mc.setup()
    >>> cat = mc.find_availability(source_cat['CIOFS']))

    Find availability for only nowcast of CBOFS model:
    >>> cat = mc.find_availability(source_cat['CBOFS'], 'nowcast')
    """

    # if no timings input, loop through all
    timings = list(cat) if timings is None else mc.astype(timings, list)

    for timing in timings:
        # check if start and end datetime files already exist and are new enough to use
        # check if already know the time and not stale
        # file times are given in UTC
        # If files are not stale, read in info from there
        if not override and mc.is_fresh(mc.FILE_PATH_START(cat.name, timing)):
            with open(mc.FILE_PATH_START(cat.name, timing), "r") as stream:
                start_datetime = yaml.safe_load(stream)["start_datetime"]
            find_start_datetime = False
        else:
            find_start_datetime = True  # need to still find the start_datetime

        if not override and mc.is_fresh(mc.FILE_PATH_END(cat.name, timing)):
            with open(mc.FILE_PATH_END(cat.name, timing), "r") as stream:
                end_datetime = yaml.safe_load(stream)["end_datetime"]
            find_end_datetime = False
        else:
            find_end_datetime = True  # need to still find the end_datetime

        # start and end temp could be None, depending on values of
        # find_start_datetime, find_end_datetime
        if find_start_datetime or find_end_datetime:
            start_temp, end_temp = find_datetimes(
                cat[timing], find_start_datetime, find_end_datetime, override=override
            )

        start_datetime = start_temp if find_start_datetime else start_datetime
        end_datetime = end_temp if find_end_datetime else end_datetime

        cat[timing].metadata["start_datetime"] = start_datetime
        cat[timing].metadata["end_datetime"] = end_datetime

    # Make new catalog to remember the new metadata
    new_user_cat = mc.make_catalog(
        [cat[timing] for timing in list(cat)],
        full_cat_name=cat.name,
        full_cat_description=cat.description,
        full_cat_metadata=cat.metadata,
        cat_driver=[cat[timing]._entry._driver for timing in list(cat)],
        cat_path=None,
        save_catalog=False,
    )

    return new_user_cat


def transform_source(source_orig):
    """Set up transform of original catalog source

    Parameters
    ----------
    source_orig : Intake source
        Original source, which will be transformed

    Returns
    -------
    source_transform, the transformed version of source_orig. This source will point at the source of
    source_orig as the target.
    """

    # open the skeleton transform cat entry and then alter
    # a few things so can use it with source_orig
    source_transform = intake.open_catalog(mc.CAT_PATH_TRANSFORM)["name"]

    # Update name and description for transformed source
    source_transform.name = source_orig.name
    source_transform.description = (
        f"Catalog entry for transform of dataset {source_orig.name}"
    )

    # make path to source_orig the target
    source_transform.__dict__["_captured_init_kwargs"]["targets"] = [
        f"{source_orig.cat.path}:{source_orig.name}"
    ]

    # add metadata from source_orig
    source_transform.metadata.update(source_orig.metadata)

    # add yesterday if needed (some RTOFS models)
    if any(
        ["yesterday" in d.values() for d in source_orig.describe()["user_parameters"]]
    ):
        yesterday = pd.Timestamp.today() - pd.Timedelta("1 day")
        # import pdb; pdb.set_trace()
        source_transform.__dict__["_captured_init_kwargs"]["transform_kwargs"][
            "yesterday"
        ] = str(yesterday)[:10]

    return source_transform


def select_date_range(
    cat, start_date, end_date, timing=None, forecast_forward=False, override=False
):
    """Add urlpath locations to existing catalog/source.

    NOAA OFS model-timings that require aggregation need to have the specific file paths found for each
    file that will be read in. This function does that, based on the desired date range, and returns a
    Source with them in the `urlpath`. This function can be also used with any model that does not
    require this (because the model paths are either static or deterministic) but in those cases it does
    not need to be used; they will have the start and end dates applied to filter the resulting model
    output after `to_dask()` is called.

    Parameters
    ----------
    cat: Intake catalog
        An intake catalog for a specific model entry.
    timing: str, optional
        Which timing to use. If `find_availability` has been run, the code will determine whether
        `start_date`, `end_date` are in "forecast" or "hindcast". Otherwise timing must be provided for a
        single timing. Normally the options are "forecast", "nowcast", or "hindcast", and sometimes
        "hindcast-forecast-aggregation". An exception is if there is only one timing available for cat,
        that one will be used without specifying it.
    start_date, end_date: datetime-interpretable str or pd.Timestamp
        For models that require aggregation, the end_date is inclusive. For all models, the start and end
        date are used with `xarray` directly with `ds.cf.sel(T=slice(start_date, end_date))` in
        `to_dask()`.
    forecast_forward : bool, optional
        Nowcast files are aggregated for the dates in the user-defined date range. However, if
        `forecast_forward==True`, the final date can have forecast files aggregated after the nowcast
        files to create a forecast going forward in time from the end date. The default is to not include
        the forecast on the end (`forecast_forward=False`).
    override : boolean, optional
        Use `override=True` to find catrefs regardless of freshness.

    Returns
    -------
    Source associated with the catalog entry.

    Examples
    --------

    Find model 'LMHOFS' urlpaths directly from source catalog without first
    search for availability with `find_availability()`:
    >>> source_cat = mc.setup_source_catalog()
    >>> today = pd.Timestamp.today()
    >>> cat = source_cat["LMHOFS"]
    >>> source = mc.add_url_path(cat, timing="forecast",
                                 start_date=today, end_date=today)

    Find availability for model (for forecast and hindcast timings), then find
    urlpaths:
    >>> cat = mc.find_availability(model='LMHOFS')
    >>> today = pd.Timestamp.today()
    >>> source = mc.add_url_path(cat, start_date=today, end_date=today)

    """

    end_date = mc.astype(end_date, pd.Timestamp)

    # if there is only one timing, use it
    if timing is None and len(list(cat)) == 1:
        timing = list(cat)[0]

    elif timing is None and any(
        [
            "start_datetime" not in cat[timing].metadata
            or "end_datetime" not in cat[timing].metadata
            for timing in list(cat)
        ]
    ):
        raise KeyError(
            "Run `mc.find_availability()` for this model before running this command. Otherwise input timing that contains desired date range."  # noqa: E501
        )

    # which source to use from catalog for desired date range
    if timing is None:
        user_range = DateTimeRange(start_date, end_date)

        for timing in list(cat):
            timing_range = DateTimeRange(
                cat[timing].metadata["start_datetime"],
                cat[timing].metadata["end_datetime"],
            )
            try:  # use this timing if it is in the date range
                if user_range in timing_range:
                    break
            except TypeError:
                continue
        else:
            raise ValueError("date range does not fully fit into any model timings")

    if forecast_forward and timing == "hindcast":
        raise KeyError(
            "timing 'hindcast' does not have forecast files, so `forecast_forward` should be False."
        )

    source = cat[timing]

    # catch the models that require aggregation
    if "catloc" in source.metadata:
        pattern = source.metadata["pattern"] if "pattern" in source.metadata else None

        if not override and mc.is_fresh(
            mc.FILE_PATH_CATREFS(source.cat.name, source.name)
        ):
            with open(
                mc.FILE_PATH_CATREFS(source.cat.name, source.name), "r"
            ) as stream:
                catrefs = yaml.safe_load(stream)["catrefs"]
        else:
            catrefs = mc.find_catrefs(source.metadata["catloc"])
            catrefs = sorted(catrefs)  # earliest first, most recent last
            with open(
                mc.FILE_PATH_CATREFS(source.cat.name, source.name), "w"
            ) as outfile:
                yaml.dump({"catrefs": catrefs}, outfile, default_flow_style=False)

        # For forecast OFS aggregations, the latest date available by catalog day
        # is today, and the forecast files within today extend forward in time.
        # in that case, use today as end date in loop below, and use
        # `forecast_forward=True`.
        today = pd.Timestamp.today()
        if end_date > today:
            end_date_use = today
            forecast_forward = True
        else:
            end_date_use = end_date

        # loop over dates
        filelocs_urlpath = []
        for date in pd.date_range(start=start_date, end=end_date_use, freq="1D"):
            is_forecast = (
                True
                if date.date() == end_date_use.date() and forecast_forward
                else False
            )

            fname = mc.FILE_PATH_AGG_FILE_LOCS(
                source.cat.name, timing, date, is_forecast
            )

            if not override and mc.is_fresh(fname):
                with open(fname, "r") as stream:
                    agg_filelocs = yaml.safe_load(stream)["agg_filelocs"]

            else:
                # translate date to catrefs to select which catref to use
                if len(catrefs[0]) == 3:
                    cat_ref_to_match = (
                        date.strftime("%Y"),
                        date.strftime("%m"),
                        date.strftime("%d"),
                    )
                elif len(catrefs[0]) == 2:
                    cat_ref_to_match = (date.strftime("%Y"), date.strftime("%m"))

                ind = catrefs.index(cat_ref_to_match)

                filelocs = mc.find_filelocs(
                    catrefs[ind], source.metadata["catloc"], cat.metadata["filetype"]
                )

                agg_filelocs = mc.agg_for_date(
                    date, filelocs, cat.metadata["filetype"], is_forecast, pattern
                )

                with open(fname, "w") as outfile:
                    yaml.dump(
                        {"agg_filelocs": agg_filelocs},
                        outfile,
                        default_flow_style=False,
                    )

            filelocs_urlpath.extend(agg_filelocs)

        # This is how we input the newly found urlpaths in so they will be used
        # in the processing of the dataset, and overwrite the old urlpath
        source.__dict__["_captured_init_kwargs"]["transform_kwargs"][
            "urlpath"
        ] = filelocs_urlpath

    # Pass start and end dates to the transform so they can be implemented
    # there for static and deterministic model files (includes RTOFS) as well
    # as the OFS aggregated models.
    source.__dict__["_captured_init_kwargs"]["transform_kwargs"]["start_date"] = str(
        start_date
    )
    source.__dict__["_captured_init_kwargs"]["transform_kwargs"]["end_date"] = str(
        end_date
    )

    # store info in source_orig
    metadata = {
        "timing": timing,
        "start_date": start_date,
        "end_date": end_date,
    }
    source.metadata.update(metadata)
    # Add original overall model catalog metadata to this next version
    source.metadata.update(cat.metadata)

    return source



"""
Make sure catalog creation is working.

Some tests won't consistently run, I think because of connection issues with
NOAA servers. I can't make progress on these so I will save them at the bottom
of this script, commented out, and I will run abbreviated sample versions of
them instead.
"""

import warnings

import pandas as pd
import pytest
import yaml

import model_catalogs as mc


def test_setup():
    """Make sure main catalog is created correctly."""

    main_cat = mc.setup(override=True)

    # check that all compiled catalog files exist
    for cat_loc in mc.CAT_PATH_ORIG.glob("*.yaml"):
        fname = mc.FILE_PATH_COMPILED(cat_loc.name)
        assert fname.exists()
        assert mc.is_fresh(fname)

    # Check that timings are correct for one test case
    assert sorted(list(main_cat["CBOFS"])) == ["forecast", "hindcast", "nowcast"]
    assert main_cat["CBOFS"].metadata["geospatial_bounds"]


@pytest.mark.slow
def test_find_availability():
    """Test find_availability.

    Using warnings instead of exception since sometimes servers aren't reliable.
    """

    # test models with fast static links and that require aggregations
    test_models = {
        "HYCOM": "forecast",
        "DBOFS": "nowcast",
        "SFBOFS": "hindcast",
    }

    main_cat = mc.setup()
    for model, timing in test_models.items():
        cat = mc.find_availability(main_cat[model], timings=timing)
        if "start_datetime" not in cat[timing].metadata:
            warnings.warn(
                f"Running model {model} with timing {timing} in `find_availability()` did not result in `start_datetime` in the catalog metadata.",  # noqa: E501
                RuntimeWarning,
            )
        fname = mc.FILE_PATH_START(model, timing)
        if not mc.is_fresh(fname):
            warnings.warn(f"Filename {fname} is not found as fresh.", RuntimeWarning)


@pytest.mark.slow
def test_boundaries():
    """Test one faster model and compare with existing file."""

    model = "hycom"

    # Calculate
    boundaries = mc.calculate_boundaries(
        file_locs=mc.FILE_PATH_ORIG(model), save_files=False, return_boundaries=True
    )

    # Read in saved
    with open(mc.FILE_PATH_BOUNDARIES(model), "r") as stream:
        boundaries_read = yaml.safe_load(stream)

    assert boundaries[model]["bbox"] == boundaries_read["bbox"]
    assert boundaries[model]["wkt"] == boundaries_read["wkt"]


@pytest.mark.slow
def test_select_date_range():
    """Test functionality in `select_date_range()`.

    Should filter date range for static and nonstatic links.
    """

    test_models = {"HYCOM": "forecast", "CIOFS": "nowcast"}

    today = pd.Timestamp.today(tz="UTC") - pd.Timedelta("1 day")
    tom = today + pd.Timedelta("1 day")

    main_cat = mc.setup()
    for model, timing in test_models.items():
        source = mc.select_date_range(
            main_cat[model], today.date(), tom.date(), timing=timing
        )

        try:
            ds = source.to_dask()
            assert (
                pd.Timestamp(ds.cf["T"].cf.isel(T=0).values).tz_localize("UTC").date()
                == today.date()
            )
            assert (
                pd.Timestamp(ds.cf["T"].cf.isel(T=-1).values).tz_localize("UTC").date()
                == tom.date()
            )

        except OSError:
            warnings.warn(
                f"Running model {model} with timing {timing} in `select_date_range()` did not return the correct date range.",  # noqa: E501
                RuntimeWarning,
            )


@pytest.mark.slow
def test_process():
    """Test that dataset is processed."""

    main_cat = mc.setup()

    # if this dataset hasn't been processed, lon and lat won't be in coords
    assert "lon" in main_cat["LOOFS"]["nowcast"].to_dask().coords


def check_source(source):
    """Check attributes of source for other tests."""

    try:
        ds = source.to_dask()
    except OSError:
        warnings.warn(
            f"Model {source.cat.name} with timing {source.name} is not working right now.",
            RuntimeWarning,
        )
        return

    # check axis attributes have been assigned
    checks = [
        ds[var_name].attrs["axis"] == axis
        for axis, var_names in source.metadata["axis"].items()
        for var_name in mc.astype(var_names, list)
        if var_name in ds.dims
    ]
    assert all(checks)

    # check standard_names attributes have been assigned
    checks = [
        ds[var_name].attrs["standard_name"] == st_name
        for st_name, var_names in source.metadata["standard_names"].items()
        for var_name in mc.astype(var_names, list)
        if var_name in ds.data_vars
    ]
    assert all(checks)

    # check cf-xarray
    # AXIS X and Y won't be defined for unstructured model unless interpolated
    if "SELFE" in source.cat.description or "FVCOM" in source.cat.description:
        if "REGULARGRID" in source.cat.name:
            assert sorted(list(ds.cf.axes.keys())) == ["T", "X", "Y", "Z"]
        elif "2DS" in source.cat.name:
            assert sorted(list(ds.cf.axes.keys())) == ["T"]
        else:
            assert sorted(list(ds.cf.axes.keys())) == ["T", "Z"]
    else:
        assert sorted(list(ds.cf.axes.keys())) == ["T", "X", "Y", "Z"]

    # the 2D cases are weird
    if "NGOFS2-2DS" in source.cat.name:
        assert sorted(list(ds.cf.coordinates.keys())) == [
            "latitude",
            "longitude",
            "time",
        ]
    else:
        assert sorted(list(ds.cf.coordinates.keys())) == [
            "latitude",
            "longitude",
            "time",
            "vertical",
        ]
    ds.close()


@pytest.mark.slow
def test_forecast():
    """Test all known models for running in forecast mode.

    Fails gracefully. Does not require running `select_date_range()` because forecasts always have some
    known files included or a static link.
    """

    main_cat = mc.setup()
    timing = "forecast"

    for cat_loc in mc.CAT_PATH_ORIG.glob("*.yaml"):
        model = cat_loc.stem.upper()
        source = main_cat[model][timing]
        try:
            check_source(source)
        except AssertionError:
            warnings.warn(
                f"Model {model} with timing {timing} does not have proper attributes.",
                RuntimeWarning,
            )


@pytest.mark.slow
def test_nowcast():
    """Test all known models for running in nowcast mode.

    Fails gracefully. Does not require running `select_date_range()` because they always have some known
    files included or a static link.
    """

    main_cat = mc.setup()
    timing = "nowcast"

    for cat_loc in mc.CAT_PATH_ORIG.glob("*.yaml"):
        model = cat_loc.stem.upper()
        if timing in list(main_cat[model]):
            source = main_cat[model][timing]
            try:
                check_source(source)
            except OSError:
                warnings.warn(
                    f"Model {model} with timing {timing} is not working right now.",
                    RuntimeWarning,
                )


@pytest.mark.slow
def test_hindcast():
    """Test all known models for running in hindcast mode."""

    main_cat = mc.setup()
    timing = "hindcast"

    for cat_loc in mc.CAT_PATH_ORIG.glob("*.yaml"):
        model = cat_loc.stem.upper()
        if timing in list(main_cat[model]):
            source = main_cat[model][timing]
            try:
                check_source(source)
            except OSError:
                warnings.warn(
                    f"Model {model} with timing {timing} is not working right now.",
                    RuntimeWarning,
                )


@pytest.mark.slow
def test_hindcast_forecast_aggregation():
    """Test all known models for running in hindcast aggregation mode."""

    main_cat = mc.setup()
    timing = "hindcast-forecast-aggregation"

    for cat_loc in mc.CAT_PATH_ORIG.glob("*.yaml"):
        model = cat_loc.stem.upper()
        if timing in list(main_cat[model]):
            source = main_cat[model][timing]
            try:
                check_source(source)
            except OSError:
                warnings.warn(
                    f"Model {model} with timing {timing} is not working right now.",
                    RuntimeWarning,
                )



"""
This file contains all information for transforming the Datasets.
"""
from typing import Optional

import cf_xarray  # noqa
import numpy as np
import xarray as xr

from intake.source.derived import GenericTransform


class DatasetTransform(GenericTransform):
    """Transform where the input and output are both Dask-compatible Datasets

    This derives from GenericTransform, and you must supply ``transform`` and
    any ``transform_kwargs``.
    """

    input_container = "xarray"
    container = "xarray"
    optional_params = {}
    _ds = None

    def to_dask(self):
        """Makes it so can read in model output."""
        if self._ds is None:
            self._pick()
            kwargs = self._params["transform_kwargs"]

            # check for "yesterday" in kwargs and use it if present
            # this is for some RTOFS models
            if "yesterday" in kwargs:
                self._source = self._source(yesterday=kwargs["yesterday"])

            # check for 'urlpath' update being sent in, if so use it to update
            if "urlpath" in kwargs:
                self._source.urlpath = kwargs["urlpath"]

            # Checks to catch potential user pitfalls
            # Make sure that user has filled in urlpath if needed: OFS nowcast
            # and some OFS forecast. These are unaggregated and come with 2
            # sample files but user should be warned if they haven't been
            # replaced since it might be a mistake
            if "sample_locs" in self._source.metadata:
                if self._source.urlpath == self._source.metadata["sample_locs"]:
                    # CHANGE TO LOGGER WARNING and also print warning
                    print(
                        "Note that you are using the original example files in your input source. You may want to instead first run `mc.select_date_range()` to search for and add the correct model output files for your desired date range, then run `to_dask()`."  # noqa: E501
                    )

            # Make sure that user has filled in urlpath if needed: OFS hindcast
            # check for if the urlpath is null and if so `select_date_range()`
            # needs to be run to fill it in
            elif self._source.urlpath is None:
                raise KeyError(
                    "The input source `urlpath` does not have a value. You probably want to run `mc.select_date_range()` before running `to_dask()`."  # noqa: E501
                )

            # This sends the metadata to `add_attributes()`
            self._ds = self._transform(
                self._source.to_dask(),
                metadata=self.metadata,
            )

            # check for 'urlpath' update being sent in, if so use it to
            # subselect ds in time
            if "start_date" in kwargs and "end_date" in kwargs:
                self._ds = self._ds.cf.sel(
                    T=slice(kwargs["start_date"], kwargs["end_date"])
                )

        return self._ds

    def read(self):
        """Same here."""
        return self.to_dask()


def add_attributes(ds, metadata: Optional[dict] = None):
    """Update Dataset metadata.

    Update the Dataset metadata with metadata passed in from catalog files.

    Parameters
    ----------
    ds : Dataset
        xarray Dataset containing model output.
    metadata : dict, optional
        Metadata that has processing information to apply

    Returns
    -------
    Improved Dataset.
    """
    # set standard_names for all variables
    if metadata is not None and "standard_names" in metadata:
        for stan_name, var_names in metadata["standard_names"].items():
            if not isinstance(var_names, list):
                var_names = [var_names]
            for var_name in var_names:
                if var_name in ds.data_vars:
                    ds[var_name].attrs["standard_name"] = stan_name

    # # Run code to find vertical coordinates
    # try:
    #     # create name mapping
    #     snames = ['ocean_s_coordinate_g1', 'ocean_s_coordinate_g2', 'ocean_sigma_coordinate']
    #     s_vars = [standard_names[sname] for sname in snames if sname in standard_names][0]
    #     z_vars = axis['Z']
    #     outnames = {s_var: z_var for s_var, z_var in zip(s_vars, z_vars)}
    #     ds.cf.decode_vertical_coords(outnames=outnames)
    # except Exception:
    #     pass

    if metadata is not None and "coords" in metadata:
        ds = ds.assign_coords({k: ds[k] for k in metadata["coords"]})

    # set axis attributes (time, lon, lat, z potentially)
    if metadata is not None and "axis" in metadata:
        for ax_name, var_names in metadata["axis"].items():
            if not isinstance(var_names, list):
                var_names = [var_names]
            for var_name in var_names:
                # var_name needs to exist
                # if ax_name == 'X':
                #     import pdb; pdb.set_trace()

                if var_name in ds.dims:
                    # var_name needs to be a coord to have attributes
                    if var_name not in ds.coords:
                        ds[var_name] = (
                            var_name,
                            np.arange(ds.sizes[var_name]),
                            {"axis": ax_name},
                        )
                    else:
                        ds[var_name].attrs["axis"] = ax_name

    # this won't run for e.g. GFS which has multiple time variables
    # but also doesn't need to have the calendar updated
    try:
        attrs = ds[ds.cf["T"].name].attrs
        if ("calendar" in attrs) and (attrs["calendar"] == "gregorian_proleptic"):
            attrs["calendar"] = "proleptic_gregorian"
            ds[ds.cf["T"].name].attrs = attrs
    except KeyError:
        pass

    # check for a dataset having both `missing_value` and `_FillValue` attributes defined
    # delete `missing_value` in this case (arbitrary decision of which to delete)
    for data_var in ds.data_vars:
        if "missing_value" in ds[data_var].attrs and "_FillValue" in ds[data_var].attrs:
            del ds[data_var].attrs["missing_value"]

    # decode times if times are floats.
    # Some datasets like GFS have multiple time coordinates for different phenomena like
    # precipitation accumulation vs winds vs surface albedo average.
    if (
        "T" in metadata["axis"]
        and isinstance(metadata["axis"]["T"], list)
        and len(metadata["axis"]["T"]) > 1
    ):
        for time_var in metadata["axis"]["T"]:
            if ds[time_var].dtype == "float64":
                ds = xr.decode_cf(ds, decode_times=True)
                break
    elif ds.cf["T"].dtype == "float64":
        ds = xr.decode_cf(ds, decode_times=True)

    if metadata is not None and "formula_terms" in metadata:
        for varname in metadata["formula_terms"]:
            ds[varname].attrs["formula_terms"] = metadata["formula_terms"][varname]

    # This is an internal attribute used by netCDF which xarray doesn't know or care about, but can
    # be returned from THREDDS.
    if "_NCProperties" in ds.attrs:
        del ds.attrs["_NCProperties"]

    return ds



"""
Setup for model_catalogs.
"""

from setuptools import setup


setup(
    use_scm_version={
        # "root": "..",  # this is because the .git dir is back a dir
        "relative_to": __file__,
        "write_to": "model_catalogs/_version.py",
        "write_to_template": '__version__ = "{version}"',
        "tag_regex": r"^(?P<prefix>v)?(?P<version>[^\+]+)(?P<suffix>.*)?$",
    }
)
